@@ -1,7 +1,6 @@
 import asyncio
 
 import engineio
-import six
 
 from . import asyncio_manager
 from . import exceptions
@@ -14,7 +13,7 @@ class AsyncServer(server.Server):
 
     This class implements a fully compliant Socket.IO web server with support
     for websocket and long-polling transports, compatible with the asyncio
-    framework on Python 3.5 or newer.
+    framework.
 
     :param client_manager: The client manager instance that will manage the
                            client list. When this is omitted, the client list
@@ -27,54 +26,82 @@ class AsyncServer(server.Server):
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
                  versions.
-    :param async_handlers: If set to ``True``, event handlers are executed in
-                           separate threads. To run handlers synchronously,
-                           set to ``False``. The default is ``True``.
+    :param async_handlers: If set to ``True``, event handlers for a client are
+                           executed in separate threads. To run handlers for a
+                           client synchronously, set to ``False``. The default
+                           is ``True``.
+    :param always_connect: When set to ``False``, new connections are
+                           provisory until the connect handler returns
+                           something other than ``False``, at which point they
+                           are accepted. When set to ``True``, connections are
+                           immediately accepted, and then if the connect
+                           handler returns ``False`` a disconnect is issued.
+                           Set to ``True`` if you need to emit events from the
+                           connect handler and your client is confused when it
+                           receives events before the connection acceptance.
+                           In any other case use the default of ``False``.
     :param kwargs: Connection parameters for the underlying Engine.IO server.
 
     The Engine.IO configuration supports the following settings:
 
     :param async_mode: The asynchronous model to use. See the Deployment
                        section in the documentation for a description of the
-                       available options. Valid async modes are "aiohttp". If
-                       this argument is not given, an async mode is chosen
-                       based on the installed packages.
+                       available options. Valid async modes are "threading",
+                       "eventlet", "gevent" and "gevent_uwsgi". If this
+                       argument is not given, "eventlet" is tried first, then
+                       "gevent_uwsgi", then "gevent", and finally "threading".
+                       The first async mode that has all its dependencies
+                       installed is then one that is chosen.
+    :param ping_interval: The interval in seconds at which the server pings
+                          the client. The default is 25 seconds. For advanced
+                          control, a two element tuple can be given, where
+                          the first number is the ping interval and the second
+                          is a grace period added by the server.
     :param ping_timeout: The time in seconds that the client waits for the
-                         server to respond before disconnecting.
-    :param ping_interval: The interval in seconds at which the client pings
-                          the server.
+                         server to respond before disconnecting. The default
+                         is 5 seconds.
     :param max_http_buffer_size: The maximum size of a message when using the
-                                 polling transport.
-    :param allow_upgrades: Whether to allow transport upgrades or not.
+                                 polling transport. The default is 1,000,000
+                                 bytes.
+    :param allow_upgrades: Whether to allow transport upgrades or not. The
+                           default is ``True``.
     :param http_compression: Whether to compress packages when using the
-                             polling transport.
+                             polling transport. The default is ``True``.
     :param compression_threshold: Only compress messages when their byte size
-                                  is greater than this value.
-    :param cookie: Name of the HTTP cookie that contains the client session
-                   id. If set to ``None``, a cookie is not sent to the client.
+                                  is greater than this value. The default is
+                                  1024 bytes.
+    :param cookie: If set to a string, it is the name of the HTTP cookie the
+                   server sends back tot he client containing the client
+                   session id. If set to a dictionary, the ``'name'`` key
+                   contains the cookie name and other keys define cookie
+                   attributes, where the value of each attribute can be a
+                   string, a callable with no arguments, or a boolean. If set
+                   to ``None`` (the default), a cookie is not sent to the
+                   client.
     :param cors_allowed_origins: Origin or list of origins that are allowed to
                                  connect to this server. Only the same origin
                                  is allowed by default. Set this argument to
                                  ``'*'`` to allow all origins, or to ``[]`` to
                                  disable CORS handling.
     :param cors_credentials: Whether credentials (cookies, authentication) are
-                             allowed in requests to this server.
+                             allowed in requests to this server. The default is
+                             ``True``.
     :param monitor_clients: If set to ``True``, a background task will ensure
                             inactive clients are closed. Set to ``False`` to
                             disable the monitoring task (not recommended). The
                             default is ``True``.
     :param engineio_logger: To enable Engine.IO logging set to ``True`` or pass
                             a logger object to use. To disable logging set to
-                            ``False``. Note that fatal errors are logged even
-                            when ``engineio_logger`` is ``False``.
+                            ``False``. The default is ``False``. Note that
+                            fatal errors are logged even when
+                            ``engineio_logger`` is ``False``.
     """
     def __init__(self, client_manager=None, logger=False, json=None,
                  async_handlers=True, **kwargs):
         if client_manager is None:
             client_manager = asyncio_manager.AsyncManager()
         super().__init__(client_manager=client_manager, logger=logger,
-                         binary=False, json=json,
-                         async_handlers=async_handlers, **kwargs)
+                         json=json, async_handlers=async_handlers, **kwargs)
 
     def is_asyncio_based(self):
         return True
@@ -91,8 +118,9 @@ class AsyncServer(server.Server):
                       ``'connect'``, ``'message'`` and ``'disconnect'`` are
                       reserved and should not be used.
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The recipient of the message. This can be set to the
                    session ID of a client to address only that client, or to
                    to any custom room created by the application to address all
@@ -142,8 +170,9 @@ class AsyncServer(server.Server):
         :func:`emit` to issue custom event names.
 
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The recipient of the message. This can be set to the
                    session ID of a client to address only that client, or to
                    to any custom room created by the application to address all
@@ -183,8 +212,9 @@ class AsyncServer(server.Server):
                       ``'connect'``, ``'message'`` and ``'disconnect'`` are
                       reserved and should not be used.
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The session ID of the recipient client.
         :param sid: Alias for the ``to`` parameter.
         :param namespace: The Socket.IO namespace for the event. If this
@@ -226,7 +256,7 @@ class AsyncServer(server.Server):
         try:
             await asyncio.wait_for(callback_event.wait(), timeout)
         except asyncio.TimeoutError:
-            six.raise_from(exceptions.TimeoutError(), None)
+            raise exceptions.TimeoutError() from None
         return callback_args[0] if len(callback_args[0]) > 1 \
             else callback_args[0][0] if len(callback_args[0]) == 1 \
             else None
@@ -258,7 +288,8 @@ class AsyncServer(server.Server):
         the user session, use the ``session`` context manager instead.
         """
         namespace = namespace or '/'
-        eio_session = await self.eio.get_session(sid)
+        eio_sid = self.manager.eio_sid_from_sid(sid, namespace)
+        eio_session = await self.eio.get_session(eio_sid)
         return eio_session.setdefault(namespace, {})
 
     async def save_session(self, sid, session, namespace=None):
@@ -270,7 +301,8 @@ class AsyncServer(server.Server):
                           the default namespace is used.
         """
         namespace = namespace or '/'
-        eio_session = await self.eio.get_session(sid)
+        eio_sid = self.manager.eio_sid_from_sid(sid, namespace)
+        eio_session = await self.eio.get_session(eio_sid)
         eio_session[namespace] = session
 
     def session(self, sid, namespace=None):
@@ -334,13 +366,11 @@ class AsyncServer(server.Server):
             delete_it = await self.manager.can_disconnect(sid, namespace)
         if delete_it:
             self.logger.info('Disconnecting %s [%s]', sid, namespace)
-            self.manager.pre_disconnect(sid, namespace=namespace)
-            await self._send_packet(sid, packet.Packet(packet.DISCONNECT,
-                                                       namespace=namespace))
+            eio_sid = self.manager.pre_disconnect(sid, namespace=namespace)
+            await self._send_packet(eio_sid, packet.Packet(
+                packet.DISCONNECT, namespace=namespace))
             await self._trigger_event('disconnect', namespace, sid)
             self.manager.disconnect(sid, namespace=namespace)
-            if namespace == '/':
-                await self.eio.disconnect(sid)
 
     async def handle_request(self, *args, **kwargs):
         """Handle an HTTP request from the client.
@@ -392,31 +422,28 @@ class AsyncServer(server.Server):
         else:
             data = []
         await self._send_packet(sid, packet.Packet(
-            packet.EVENT, namespace=namespace, data=[event] + data, id=id,
-            binary=None))
+            packet.EVENT, namespace=namespace, data=[event] + data, id=id))
 
-    async def _send_packet(self, sid, pkt):
+    async def _send_packet(self, eio_sid, pkt):
         """Send a Socket.IO packet to a client."""
         encoded_packet = pkt.encode()
         if isinstance(encoded_packet, list):
-            binary = False
             for ep in encoded_packet:
-                await self.eio.send(sid, ep, binary=binary)
-                binary = True
+                await self.eio.send(eio_sid, ep)
         else:
-            await self.eio.send(sid, encoded_packet, binary=False)
+            await self.eio.send(eio_sid, encoded_packet)
 
-    async def _handle_connect(self, sid, namespace):
+    async def _handle_connect(self, eio_sid, namespace):
         """Handle a client connection request."""
         namespace = namespace or '/'
-        self.manager.connect(sid, namespace)
+        sid = self.manager.connect(eio_sid, namespace)
         if self.always_connect:
-            await self._send_packet(sid, packet.Packet(packet.CONNECT,
-                                                       namespace=namespace))
-        fail_reason = None
+            await self._send_packet(eio_sid, packet.Packet(
+                packet.CONNECT, {'sid': sid}, namespace=namespace))
+        fail_reason = exceptions.ConnectionRefusedError().error_args
         try:
             success = await self._trigger_event('connect', namespace, sid,
-                                                self.environ[sid])
+                                                self.environ[eio_sid])
         except exceptions.ConnectionRefusedError as exc:
             fail_reason = exc.error_args
             success = False
@@ -424,39 +451,31 @@ class AsyncServer(server.Server):
         if success is False:
             if self.always_connect:
                 self.manager.pre_disconnect(sid, namespace)
-                await self._send_packet(sid, packet.Packet(
+                await self._send_packet(eio_sid, packet.Packet(
                     packet.DISCONNECT, data=fail_reason, namespace=namespace))
-            elif namespace != '/':
-                await self._send_packet(sid, packet.Packet(
-                    packet.ERROR, data=fail_reason, namespace=namespace))
+            else:
+                await self._send_packet(eio_sid, packet.Packet(
+                    packet.CONNECT_ERROR, data=fail_reason,
+                    namespace=namespace))
             self.manager.disconnect(sid, namespace)
-            if namespace == '/' and sid in self.environ:  # pragma: no cover
-                del self.environ[sid]
-            return fail_reason or False
         elif not self.always_connect:
-            await self._send_packet(sid, packet.Packet(packet.CONNECT,
-                                                       namespace=namespace))
+            await self._send_packet(eio_sid, packet.Packet(
+                packet.CONNECT, {'sid': sid}, namespace=namespace))
 
-    async def _handle_disconnect(self, sid, namespace):
+    async def _handle_disconnect(self, eio_sid, namespace):
         """Handle a client disconnect."""
         namespace = namespace or '/'
-        if namespace == '/':
-            namespace_list = list(self.manager.get_namespaces())
-        else:
-            namespace_list = [namespace]
-        for n in namespace_list:
-            if n != '/' and self.manager.is_connected(sid, n):
-                self.manager.pre_disconnect(sid, namespace=n)
-                await self._trigger_event('disconnect', n, sid)
-                self.manager.disconnect(sid, n)
-        if namespace == '/' and self.manager.is_connected(sid, namespace):
-            self.manager.pre_disconnect(sid, namespace='/')
-            await self._trigger_event('disconnect', '/', sid)
-            self.manager.disconnect(sid, '/')
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
+        if not self.manager.is_connected(sid, namespace):  # pragma: no cover
+            return
+        self.manager.pre_disconnect(sid, namespace=namespace)
+        await self._trigger_event('disconnect', namespace, sid)
+        self.manager.disconnect(sid, namespace)
 
-    async def _handle_event(self, sid, namespace, id, data):
+    async def _handle_event(self, eio_sid, namespace, id, data):
         """Handle an incoming client event."""
         namespace = namespace or '/'
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
         self.logger.info('received event "%s" from %s [%s]', data[0], sid,
                          namespace)
         if not self.manager.is_connected(sid, namespace):
@@ -465,11 +484,13 @@ class AsyncServer(server.Server):
             return
         if self.async_handlers:
             self.start_background_task(self._handle_event_internal, self, sid,
-                                       data, namespace, id)
+                                       eio_sid, data, namespace, id)
         else:
-            await self._handle_event_internal(self, sid, data, namespace, id)
+            await self._handle_event_internal(self, sid, eio_sid, data,
+                                              namespace, id)
 
-    async def _handle_event_internal(self, server, sid, data, namespace, id):
+    async def _handle_event_internal(self, server, sid, eio_sid, data,
+                                     namespace, id):
         r = await server._trigger_event(data[0], namespace, sid, *data[1:])
         if id is not None:
             # send ACK packet with the response returned by the handler
@@ -480,16 +501,15 @@ class AsyncServer(server.Server):
                 data = list(r)
             else:
                 data = [r]
-            await server._send_packet(sid, packet.Packet(packet.ACK,
-                                                         namespace=namespace,
-                                                         id=id, data=data,
-                                                         binary=None))
+            await server._send_packet(eio_sid, packet.Packet(
+                packet.ACK, namespace=namespace, id=id, data=data))
 
-    async def _handle_ack(self, sid, namespace, id, data):
+    async def _handle_ack(self, eio_sid, namespace, id, data):
         """Handle ACK packets from the client."""
         namespace = namespace or '/'
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
         self.logger.info('received ack from %s [%s]', sid, namespace)
-        await self.manager.trigger_callback(sid, namespace, id, data)
+        await self.manager.trigger_callback(sid, id, data)
 
     async def _trigger_event(self, event, namespace, *args):
         """Invoke an application event handler."""
@@ -510,49 +530,51 @@ class AsyncServer(server.Server):
             return await self.namespace_handlers[namespace].trigger_event(
                 event, *args)
 
-    async def _handle_eio_connect(self, sid, environ):
+    async def _handle_eio_connect(self, eio_sid, environ):
         """Handle the Engine.IO connection event."""
         if not self.manager_initialized:
             self.manager_initialized = True
             self.manager.initialize()
-        self.environ[sid] = environ
-        return await self._handle_connect(sid, '/')
+        self.environ[eio_sid] = environ
 
-    async def _handle_eio_message(self, sid, data):
+    async def _handle_eio_message(self, eio_sid, data):
         """Dispatch Engine.IO messages."""
-        if sid in self._binary_packet:
-            pkt = self._binary_packet[sid]
+        if eio_sid in self._binary_packet:
+            pkt = self._binary_packet[eio_sid]
             if pkt.add_attachment(data):
-                del self._binary_packet[sid]
+                del self._binary_packet[eio_sid]
                 if pkt.packet_type == packet.BINARY_EVENT:
-                    await self._handle_event(sid, pkt.namespace, pkt.id,
+                    await self._handle_event(eio_sid, pkt.namespace, pkt.id,
                                              pkt.data)
                 else:
-                    await self._handle_ack(sid, pkt.namespace, pkt.id,
+                    await self._handle_ack(eio_sid, pkt.namespace, pkt.id,
                                            pkt.data)
         else:
             pkt = packet.Packet(encoded_packet=data)
             if pkt.packet_type == packet.CONNECT:
-                await self._handle_connect(sid, pkt.namespace)
+                await self._handle_connect(eio_sid, pkt.namespace)
             elif pkt.packet_type == packet.DISCONNECT:
-                await self._handle_disconnect(sid, pkt.namespace)
+                await self._handle_disconnect(eio_sid, pkt.namespace)
             elif pkt.packet_type == packet.EVENT:
-                await self._handle_event(sid, pkt.namespace, pkt.id, pkt.data)
+                await self._handle_event(eio_sid, pkt.namespace, pkt.id,
+                                         pkt.data)
             elif pkt.packet_type == packet.ACK:
-                await self._handle_ack(sid, pkt.namespace, pkt.id, pkt.data)
+                await self._handle_ack(eio_sid, pkt.namespace, pkt.id,
+                                       pkt.data)
             elif pkt.packet_type == packet.BINARY_EVENT or \
                     pkt.packet_type == packet.BINARY_ACK:
-                self._binary_packet[sid] = pkt
-            elif pkt.packet_type == packet.ERROR:
-                raise ValueError('Unexpected ERROR packet.')
+                self._binary_packet[eio_sid] = pkt
+            elif pkt.packet_type == packet.CONNECT_ERROR:
+                raise ValueError('Unexpected CONNECT_ERROR packet.')
             else:
                 raise ValueError('Unknown packet type.')
 
-    async def _handle_eio_disconnect(self, sid):
+    async def _handle_eio_disconnect(self, eio_sid):
         """Handle Engine.IO disconnect event."""
-        await self._handle_disconnect(sid, '/')
-        if sid in self.environ:
-            del self.environ[sid]
+        for n in list(self.manager.get_namespaces()).copy():
+            await self._handle_disconnect(eio_sid, n)
+        if eio_sid in self.environ:
+            del self.environ[eio_sid]
 
     def _engineio_server_class(self):
         return engineio.AsyncServer
