@@ -1,7 +1,6 @@
 import logging
 
 import engineio
-import six
 
 from . import base_manager
 from . import exceptions
@@ -25,12 +24,6 @@ class Server(object):
                    use. To disable logging set to ``False``. The default is
                    ``False``. Note that fatal errors are logged even when
                    ``logger`` is ``False``.
-    :param binary: ``True`` to support binary payloads, ``False`` to treat all
-                   payloads as text. On Python 2, if this is set to ``True``,
-                   ``unicode`` values are treated as text, and ``str`` and
-                   ``bytes`` values are treated as binary.  This option has no
-                   effect on Python 3, where text and binary payloads are
-                   always automatically discovered.
     :param json: An alternative json module to use for encoding and decoding
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
@@ -61,13 +54,16 @@ class Server(object):
                        "gevent_uwsgi", then "gevent", and finally "threading".
                        The first async mode that has all its dependencies
                        installed is then one that is chosen.
+    :param ping_interval: The interval in seconds at which the server pings
+                          the client. The default is 25 seconds. For advanced
+                          control, a two element tuple can be given, where
+                          the first number is the ping interval and the second
+                          is a grace period added by the server.
     :param ping_timeout: The time in seconds that the client waits for the
                          server to respond before disconnecting. The default
-                         is 60 seconds.
-    :param ping_interval: The interval in seconds at which the client pings
-                          the server. The default is 25 seconds.
+                         is 5 seconds.
     :param max_http_buffer_size: The maximum size of a message when using the
-                                 polling transport. The default is 100,000,000
+                                 polling transport. The default is 1,000,000
                                  bytes.
     :param allow_upgrades: Whether to allow transport upgrades or not. The
                            default is ``True``.
@@ -76,9 +72,14 @@ class Server(object):
     :param compression_threshold: Only compress messages when their byte size
                                   is greater than this value. The default is
                                   1024 bytes.
-    :param cookie: Name of the HTTP cookie that contains the client session
-                   id. If set to ``None``, a cookie is not sent to the client.
-                   The default is ``'io'``.
+    :param cookie: If set to a string, it is the name of the HTTP cookie the
+                   server sends back tot he client containing the client
+                   session id. If set to a dictionary, the ``'name'`` key
+                   contains the cookie name and other keys define cookie
+                   attributes, where the value of each attribute can be a
+                   string, a callable with no arguments, or a boolean. If set
+                   to ``None`` (the default), a cookie is not sent to the
+                   client.
     :param cors_allowed_origins: Origin or list of origins that are allowed to
                                  connect to this server. Only the same origin
                                  is allowed by default. Set this argument to
@@ -97,9 +98,8 @@ class Server(object):
                             fatal errors are logged even when
                             ``engineio_logger`` is ``False``.
     """
-    def __init__(self, client_manager=None, logger=False, binary=False,
-                 json=None, async_handlers=True, always_connect=False,
-                 **kwargs):
+    def __init__(self, client_manager=None, logger=False, json=None,
+                 async_handlers=True, always_connect=False, **kwargs):
         engineio_options = kwargs
         engineio_logger = engineio_options.pop('engineio_logger', None)
         if engineio_logger is not None:
@@ -112,7 +112,6 @@ class Server(object):
         self.eio.on('connect', self._handle_eio_connect)
         self.eio.on('message', self._handle_eio_message)
         self.eio.on('disconnect', self._handle_eio_disconnect)
-        self.binary = binary
 
         self.environ = {}
         self.handlers = {}
@@ -124,8 +123,7 @@ class Server(object):
             self.logger = logger
         else:
             self.logger = default_logger
-            if not logging.root.handlers and \
-                    self.logger.level == logging.NOTSET:
+            if self.logger.level == logging.NOTSET:
                 if logger:
                     self.logger.setLevel(logging.INFO)
                 else:
@@ -253,8 +251,9 @@ class Server(object):
                       ``'connect'``, ``'message'`` and ``'disconnect'`` are
                       reserved and should not be used.
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The recipient of the message. This can be set to the
                    session ID of a client to address only that client, or to
                    to any custom room created by the application to address all
@@ -302,8 +301,9 @@ class Server(object):
         :func:`emit` to issue custom event names.
 
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The recipient of the message. This can be set to the
                    session ID of a client to address only that client, or to
                    to any custom room created by the application to address all
@@ -341,8 +341,9 @@ class Server(object):
                       ``'connect'``, ``'message'`` and ``'disconnect'`` are
                       reserved and should not be used.
         :param data: The data to send to the client or clients. Data can be of
-                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
-                     ``list`` or ``dict``, the data will be serialized as JSON.
+                     type ``str``, ``bytes``, ``list`` or ``dict``. To send
+                     multiple arguments, use a tuple where each element is of
+                     one of the types indicated above.
         :param to: The session ID of the recipient client.
         :param sid: Alias for the ``to`` parameter.
         :param namespace: The Socket.IO namespace for the event. If this
@@ -451,7 +452,8 @@ class Server(object):
         is used.
         """
         namespace = namespace or '/'
-        eio_session = self.eio.get_session(sid)
+        eio_sid = self.manager.eio_sid_from_sid(sid, namespace)
+        eio_session = self.eio.get_session(eio_sid)
         return eio_session.setdefault(namespace, {})
 
     def save_session(self, sid, session, namespace=None):
@@ -463,7 +465,8 @@ class Server(object):
                           the default namespace is used.
         """
         namespace = namespace or '/'
-        eio_session = self.eio.get_session(sid)
+        eio_sid = self.manager.eio_sid_from_sid(sid, namespace)
+        eio_session = self.eio.get_session(eio_sid)
         eio_session[namespace] = session
 
     def session(self, sid, namespace=None):
@@ -525,13 +528,11 @@ class Server(object):
             delete_it = self.manager.can_disconnect(sid, namespace)
         if delete_it:
             self.logger.info('Disconnecting %s [%s]', sid, namespace)
-            self.manager.pre_disconnect(sid, namespace=namespace)
-            self._send_packet(sid, packet.Packet(packet.DISCONNECT,
-                                                 namespace=namespace))
+            eio_sid = self.manager.pre_disconnect(sid, namespace=namespace)
+            self._send_packet(eio_sid, packet.Packet(
+                packet.DISCONNECT, namespace=namespace))
             self._trigger_event('disconnect', namespace, sid)
             self.manager.disconnect(sid, namespace=namespace)
-            if namespace == '/':
-                self.eio.disconnect(sid)
 
     def transport(self, sid):
         """Return the name of the transport used by the client.
@@ -543,6 +544,16 @@ class Server(object):
         """
         return self.eio.transport(sid)
 
+    def get_environ(self, sid, namespace=None):
+        """Return the WSGI environ dictionary for a client.
+
+        :param sid: The session of the client.
+        :param namespace: The Socket.IO namespace. If this argument is omitted
+                          the default namespace is used.
+        """
+        eio_sid = self.manager.eio_sid_from_sid(sid, namespace or '/')
+        return self.environ.get(eio_sid)
+
     def handle_request(self, environ, start_response):
         """Handle an HTTP request from the client.
 
@@ -586,12 +597,8 @@ class Server(object):
         """
         return self.eio.sleep(seconds)
 
-    def _emit_internal(self, sid, event, data, namespace=None, id=None):
+    def _emit_internal(self, eio_sid, event, data, namespace=None, id=None):
         """Send a message to a client."""
-        if six.PY2 and not self.binary:
-            binary = False  # pragma: nocover
-        else:
-            binary = None
         # tuples are expanded to multiple arguments, everything else is sent
         # as a single argument
         if isinstance(data, tuple):
@@ -600,32 +607,29 @@ class Server(object):
             data = [data]
         else:
             data = []
-        self._send_packet(sid, packet.Packet(packet.EVENT, namespace=namespace,
-                                             data=[event] + data, id=id,
-                                             binary=binary))
+        self._send_packet(eio_sid, packet.Packet(
+            packet.EVENT, namespace=namespace, data=[event] + data, id=id))
 
-    def _send_packet(self, sid, pkt):
+    def _send_packet(self, eio_sid, pkt):
         """Send a Socket.IO packet to a client."""
         encoded_packet = pkt.encode()
         if isinstance(encoded_packet, list):
-            binary = False
             for ep in encoded_packet:
-                self.eio.send(sid, ep, binary=binary)
-                binary = True
+                self.eio.send(eio_sid, ep)
         else:
-            self.eio.send(sid, encoded_packet, binary=False)
+            self.eio.send(eio_sid, encoded_packet)
 
-    def _handle_connect(self, sid, namespace):
+    def _handle_connect(self, eio_sid, namespace):
         """Handle a client connection request."""
         namespace = namespace or '/'
-        self.manager.connect(sid, namespace)
+        sid = self.manager.connect(eio_sid, namespace)
         if self.always_connect:
-            self._send_packet(sid, packet.Packet(packet.CONNECT,
-                                                 namespace=namespace))
-        fail_reason = None
+            self._send_packet(eio_sid, packet.Packet(
+                packet.CONNECT, {'sid': sid}, namespace=namespace))
+        fail_reason = exceptions.ConnectionRefusedError().error_args
         try:
             success = self._trigger_event('connect', namespace, sid,
-                                          self.environ[sid])
+                                          self.environ[eio_sid])
         except exceptions.ConnectionRefusedError as exc:
             fail_reason = exc.error_args
             success = False
@@ -633,39 +637,31 @@ class Server(object):
         if success is False:
             if self.always_connect:
                 self.manager.pre_disconnect(sid, namespace)
-                self._send_packet(sid, packet.Packet(
+                self._send_packet(eio_sid, packet.Packet(
                     packet.DISCONNECT, data=fail_reason, namespace=namespace))
-            elif namespace != '/':
-                self._send_packet(sid, packet.Packet(
-                    packet.ERROR, data=fail_reason, namespace=namespace))
+            else:
+                self._send_packet(eio_sid, packet.Packet(
+                    packet.CONNECT_ERROR, data=fail_reason,
+                    namespace=namespace))
             self.manager.disconnect(sid, namespace)
-            if namespace == '/' and sid in self.environ:  # pragma: no cover
-                del self.environ[sid]
-            return fail_reason or False
         elif not self.always_connect:
-            self._send_packet(sid, packet.Packet(packet.CONNECT,
-                                                 namespace=namespace))
+            self._send_packet(eio_sid, packet.Packet(
+                packet.CONNECT, {'sid': sid}, namespace=namespace))
 
-    def _handle_disconnect(self, sid, namespace):
+    def _handle_disconnect(self, eio_sid, namespace):
         """Handle a client disconnect."""
         namespace = namespace or '/'
-        if namespace == '/':
-            namespace_list = list(self.manager.get_namespaces())
-        else:
-            namespace_list = [namespace]
-        for n in namespace_list:
-            if n != '/' and self.manager.is_connected(sid, n):
-                self.manager.pre_disconnect(sid, namespace=n)
-                self._trigger_event('disconnect', n, sid)
-                self.manager.disconnect(sid, n)
-        if namespace == '/' and self.manager.is_connected(sid, namespace):
-            self.manager.pre_disconnect(sid, namespace='/')
-            self._trigger_event('disconnect', '/', sid)
-            self.manager.disconnect(sid, '/')
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
+        if not self.manager.is_connected(sid, namespace):  # pragma: no cover
+            return
+        self.manager.pre_disconnect(sid, namespace=namespace)
+        self._trigger_event('disconnect', namespace, sid)
+        self.manager.disconnect(sid, namespace)
 
-    def _handle_event(self, sid, namespace, id, data):
+    def _handle_event(self, eio_sid, namespace, id, data):
         """Handle an incoming client event."""
         namespace = namespace or '/'
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
         self.logger.info('received event "%s" from %s [%s]', data[0], sid,
                          namespace)
         if not self.manager.is_connected(sid, namespace):
@@ -674,11 +670,13 @@ class Server(object):
             return
         if self.async_handlers:
             self.start_background_task(self._handle_event_internal, self, sid,
-                                       data, namespace, id)
+                                       eio_sid, data, namespace, id)
         else:
-            self._handle_event_internal(self, sid, data, namespace, id)
+            self._handle_event_internal(self, sid, eio_sid, data, namespace,
+                                        id)
 
-    def _handle_event_internal(self, server, sid, data, namespace, id):
+    def _handle_event_internal(self, server, sid, eio_sid, data, namespace,
+                               id):
         r = server._trigger_event(data[0], namespace, sid, *data[1:])
         if id is not None:
             # send ACK packet with the response returned by the handler
@@ -689,20 +687,15 @@ class Server(object):
                 data = list(r)
             else:
                 data = [r]
-            if six.PY2 and not self.binary:
-                binary = False  # pragma: nocover
-            else:
-                binary = None
-            server._send_packet(sid, packet.Packet(packet.ACK,
-                                                   namespace=namespace,
-                                                   id=id, data=data,
-                                                   binary=binary))
+            server._send_packet(eio_sid, packet.Packet(
+                packet.ACK, namespace=namespace, id=id, data=data))
 
-    def _handle_ack(self, sid, namespace, id, data):
+    def _handle_ack(self, eio_sid, namespace, id, data):
         """Handle ACK packets from the client."""
         namespace = namespace or '/'
+        sid = self.manager.sid_from_eio_sid(eio_sid, namespace)
         self.logger.info('received ack from %s [%s]', sid, namespace)
-        self.manager.trigger_callback(sid, namespace, id, data)
+        self.manager.trigger_callback(sid, id, data)
 
     def _trigger_event(self, event, namespace, *args):
         """Invoke an application event handler."""
@@ -715,47 +708,48 @@ class Server(object):
             return self.namespace_handlers[namespace].trigger_event(
                 event, *args)
 
-    def _handle_eio_connect(self, sid, environ):
+    def _handle_eio_connect(self, eio_sid, environ):
         """Handle the Engine.IO connection event."""
         if not self.manager_initialized:
             self.manager_initialized = True
             self.manager.initialize()
-        self.environ[sid] = environ
-        return self._handle_connect(sid, '/')
+        self.environ[eio_sid] = environ
 
-    def _handle_eio_message(self, sid, data):
+    def _handle_eio_message(self, eio_sid, data):
         """Dispatch Engine.IO messages."""
-        if sid in self._binary_packet:
-            pkt = self._binary_packet[sid]
+        if eio_sid in self._binary_packet:
+            pkt = self._binary_packet[eio_sid]
             if pkt.add_attachment(data):
-                del self._binary_packet[sid]
+                del self._binary_packet[eio_sid]
                 if pkt.packet_type == packet.BINARY_EVENT:
-                    self._handle_event(sid, pkt.namespace, pkt.id, pkt.data)
+                    self._handle_event(eio_sid, pkt.namespace, pkt.id,
+                                       pkt.data)
                 else:
-                    self._handle_ack(sid, pkt.namespace, pkt.id, pkt.data)
+                    self._handle_ack(eio_sid, pkt.namespace, pkt.id, pkt.data)
         else:
             pkt = packet.Packet(encoded_packet=data)
             if pkt.packet_type == packet.CONNECT:
-                self._handle_connect(sid, pkt.namespace)
+                self._handle_connect(eio_sid, pkt.namespace)
             elif pkt.packet_type == packet.DISCONNECT:
-                self._handle_disconnect(sid, pkt.namespace)
+                self._handle_disconnect(eio_sid, pkt.namespace)
             elif pkt.packet_type == packet.EVENT:
-                self._handle_event(sid, pkt.namespace, pkt.id, pkt.data)
+                self._handle_event(eio_sid, pkt.namespace, pkt.id, pkt.data)
             elif pkt.packet_type == packet.ACK:
-                self._handle_ack(sid, pkt.namespace, pkt.id, pkt.data)
+                self._handle_ack(eio_sid, pkt.namespace, pkt.id, pkt.data)
             elif pkt.packet_type == packet.BINARY_EVENT or \
                     pkt.packet_type == packet.BINARY_ACK:
-                self._binary_packet[sid] = pkt
-            elif pkt.packet_type == packet.ERROR:
-                raise ValueError('Unexpected ERROR packet.')
+                self._binary_packet[eio_sid] = pkt
+            elif pkt.packet_type == packet.CONNECT_ERROR:
+                raise ValueError('Unexpected CONNECT_ERROR packet.')
             else:
                 raise ValueError('Unknown packet type.')
 
-    def _handle_eio_disconnect(self, sid):
+    def _handle_eio_disconnect(self, eio_sid):
         """Handle Engine.IO disconnect event."""
-        self._handle_disconnect(sid, '/')
-        if sid in self.environ:
-            del self.environ[sid]
+        for n in list(self.manager.get_namespaces()).copy():
+            self._handle_disconnect(eio_sid, n)
+        if eio_sid in self.environ:
+            del self.environ[eio_sid]
 
     def _engineio_server_class(self):
         return engineio.Server
